import {
  IMiddleware,
  Middleware,
  NextFunction,
  MidwayWebRouterService,
  Inject,
  Config,
} from '@midwayjs/core';
import { JwtService } from '@midwayjs/jwt';
import { Context } from '@midwayjs/koa';
import { get } from 'lodash';
import { LoginRecordService } from '../service/frame/LoginRecord.service';
import { ErrorCode } from '../error/Codes';
import { ServiceError } from '../error/CustomerError.error';

@Middleware()
export class CheckTokenMiddleware
  implements IMiddleware<Context, NextFunction>
{
  @Config('useCheckTokenMiddleware')
  useCheckTokenMiddleware: boolean;
  @Config('passToken')
  passTokenList: Array<string>;
  @Config('tokenKey')
  tokenKey: string;
  @Config('koa.globalPrefix')
  globalPrefix: string;
  @Inject()
  webRouterService: MidwayWebRouterService;
  @Inject()
  jwtService: JwtService;

  /** 获取tokenKey,检查header中的token是否存在,并解析出来,放到user中 */
  async analysisToken(ctx: Context) {
    const token: any = get(ctx.request, `header.${this.tokenKey}`);
    if (token) {
      const data = this.jwtService.decodeSync(token);
      if (data) {
        ctx.token = token;
        ctx.user = data;
      }
    }
    return token;
  }
  /** 排除全局前缀的路由路径 */
  async excludePrefixPath(ctx: Context) {
    // 获取路由
    const routeInfo = await this.webRouterService.getMatchedRouterInfo(
      ctx.path,
      ctx.method
    );
    const fullUrl = routeInfo.fullUrl;
    return fullUrl.replace(this.globalPrefix, '');
  }
  resolve() {
    return async (ctx: Context, next: NextFunction) => {
      let token = null;
      if (!this.useCheckTokenMiddleware) {
        await next();
      } else {
        // 解析token
        token = await this.analysisToken(ctx);
        const path = await this.excludePrefixPath(ctx);
        // 检查路由是否在 无需检查token列表中
        if (this.passTokenList && this.passTokenList.includes(path)) {
          // 在跳过token检查中,不进行处理
          await next();
        } else {
          // 检查token:是否可用
          if (!token) {
            throw new ServiceError(ErrorCode.NOT_LOGIN);
          }
          // 需要到LoginRecord表中检查:检查过期时间,token值是否一致
          const loginRecordService = await ctx.requestContext.getAsync(
            LoginRecordService
          );
          // 检查,不通过会报异常
          await loginRecordService.check(token);
          // 续期
          await loginRecordService.renewal(token);
          // 进入后续程序
          await next();
        }
      }
    };
  }
}
